/* @(#)strctrl.c 1.71 Software Membership 12/8/95 */
/*
	SM/Editor - Version 2

    ---------
    Copyright (C)  Si-Won Cho
            Dept. of Electrical Engineering
            Dongguk University
            Seoul, KOREA
            chosw@dgu4680.dongguk.ac.kr

            SAMSUNG Electronics
            Software Membership

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    ---------
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef hpux
#include <strings.h>
#endif
#include <X11/Intrinsic.h>
#include <Xm/Text.h>
#if (XmVersion < 1002)
#include <Xm/TextSrcP.h>
#endif
#include <Xm/TextP.h>
#include <Xm/TextOutP.h>	/* only for NOLINE constant */
#define SHIFT_RIGHT     0
#define SHIFT_LEFT      1

Boolean ReplaceText;

#ifdef _NO_PROTO
void parseFilename(FullName, PathName, FileName)
char *FullName;
char **PathName;
char **FileName;
#else
void parseFilename(char *FullName, char **PathName, char **FileName)
#endif
{
        int i, pathlen, filelen;
        int len = strlen(FullName);

        /* find the last slash */
        for (i = len - 1; i >= 0; i--)
    	    if (FullName[i] == '/')
	        break;

        pathlen = i + 1;
        filelen = len - pathlen;

        *PathName = XtMalloc((Cardinal)pathlen + 1);
        strncpy(*PathName, FullName, pathlen);
        *(*PathName + pathlen) = '\0';

        *FileName = XtMalloc((Cardinal)filelen + 1);
        strncpy(*FileName, FullName + pathlen, filelen);
        *(*FileName + filelen) = '\0';
}

#ifdef _NO_PROTO
void SetTabSize(w, TabSize)
Widget w;
int TabSize;
#else
void SetTabSize(Widget w, int TabSize)
#endif
{
        OutputData data = ((XmTextWidget)w)->text.output->data;

        data->tabwidth = TabSize * data->averagecharwidth;
}

#ifdef _NO_PROTO
void smTextReplace(w, start, end, str)
Widget w;
XmTextPosition start;
XmTextPosition end;
char *str;
#else
void smTextReplace(Widget w, XmTextPosition start, XmTextPosition end, char *str)
#endif
{
        ReplaceText = True;
        XmTextReplace(w, start, end, str);
        ReplaceText = False;
}

#ifdef _NO_PROTO
int textPosToLineNum(w, position)
Widget w;
XmTextPosition position;
#else
int textPosToLineNum(Widget w, XmTextPosition position)
#endif
{
        int lineNum;
        Boolean found = True;
        XmTextPosition pos;

        lineNum = _XmTextPosToLine((XmTextWidget)w, position);
        if (lineNum == NOLINE) {
            pos = XmTextGetLastPosition(w) >> 1;
            if (pos >= position) {
                pos = 0;
                lineNum = 1;
                while (found && pos < position) {
                    found = XmTextFindString(w, pos, "\n", XmTEXT_FORWARD, &pos);
                    if (found) {
                        pos++;
                        lineNum++;
                    }
                }
                return (lineNum);
            }
            else {
                pos = XmTextGetLastPosition(w);
                lineNum = ((XmTextWidget)w)->text.total_lines + 1;
                while (found && pos > position) {
                    found = XmTextFindString(w, pos - 1, "\n", XmTEXT_BACKWARD, &pos);
                    if (found) lineNum--;
                    else lineNum = 1;
                }
                return (lineNum);
            }
        }
        return (((XmTextWidget)w)->text.top_line + lineNum + 1);
}

#ifdef _NO_PROTO
void smPosToLineColumn(w, position, tab_size, line, column)
Widget w;
XmTextPosition position;
int tab_size;
int *line;
int *column;
#else         /* _NO_PROTO */
void smPosToLineColumn(Widget w, XmTextPosition position, int tab_size, int *line, int *column)
#endif
{
	int col_count, i;
/*	int tab_size = 8;*/

#if (XmVersion >= 1002)
        int status;
        char *buffer;
        Boolean found;
        XmTextPosition pos;

        if (position == 0)
            *line = 1;
        *column = 1;
        if (position > 0L) {
            *line = textPosToLineNum(w, position);
            found = XmTextFindString(w, position - 1, "\n", XmTEXT_BACKWARD, &pos);
            if (found) pos++;
            else pos = 0;
            col_count = 1;
            if (position > pos) {
                buffer = XtMalloc((Cardinal)(position - pos + 1));
                status = XmTextGetSubstring(w, pos, (int)(position - pos),
                        (int)(position - pos + 1), buffer);
	        if (status != XmCOPY_FAILED)
                    for (i = 0; i < position - pos; i++) {
	                if (buffer[i] == '\t')
	                    col_count = ((col_count + tab_size - 1) / tab_size) * tab_size + 1;
                        else col_count++;
                    }
	        XtFree(buffer);
	    }
	    *column = col_count;
        }
#else
        int i;
	int last_line = 0;
	char *str;

	str = XmTextGetString(w);
	col_count = 1;
	*line = 1;
	*column = 1;
	if (position > 0L || XmTextGetLastPosition(w) > 0L) {
            col_count = 1;
	    for (i = 0; i < (int)position; i++)
	        if (str[i] == '\n') {
	            (*line)++;
	            last_line = i + 1;
	        }
	    for (i = last_line; i < (int)position; i++)
	        if (str[i] == '\t')
	            col_count = ((col_count + tab_size - 1) / tab_size) * tab_size + 1;
	        else col_count++;
	    *column = col_count;
	}
	XtFree(str);
#endif
}

#ifdef _NO_PROTO
int smTextGetRows(w)
Widget w;
#else         /* _NO_PROTO */
int smTextGetRows(Widget w)
#endif
{
	short row;

	XtVaGetValues(w,
	        XmNrows, &row,
	        NULL);

	return ((int)(row - 1));
}

#ifdef _NO_PROTO
XmTextPosition smTextGetBottomCharacter(w)
Widget w;
#else
XmTextPosition smTextGetBottomCharacter(Widget w)
#endif
{
#if (XmVersion >= 1002) /* Motif 1.2 */
	int page, i = 0;
        short row;
        XmTextPosition pos;
        Boolean found;

	XtVaGetValues(w,
		XmNrows, 		&row,
		XmNtopCharacter,	&pos,
		NULL);
        page = (int)row - 1;
        do {
            found = XmTextFindString(w, pos, "\n", XmTEXT_FORWARD, &pos);
            if (found) {
                i++;
                pos++;
            } /* if */
        } while (found && i < page);
#else
        int page, len, i = 0;
        XmTextPosition pos;
        short row;
        char *str;

	XtVaGetValues(w,
		XmNrows, 		&row,
		XmNvalue,		&str,
		XmNtopCharacter,	&pos,
		NULL);
        page = (int)row - 1;
        len = strlen(str);
        while (pos < len && i < page)
            if (str[pos++] == '\n') i++;
        XtFree(str);
	if (pos > len) pos = len;
#endif
        return (pos);
}

#ifdef _NO_PROTO
void smTextSetInsertionPosition(w, pos)
Widget w;
XmTextPosition pos;
#else         /* _NO_PROTO */
void smTextSetInsertionPosition(Widget w, XmTextPosition pos)
#endif
{
	int row;
	XmTextPosition top_1, top_2;

	row = smTextGetRows(w) >> 1;

	top_1 = XmTextGetTopCharacter(w);
	XmTextSetInsertionPosition(w, pos);
	top_2 = XmTextGetTopCharacter(w);

	if (top_2 < top_1) XmTextScroll(w, -row);
	else if (top_2 > top_1) XmTextScroll(w, row);
}

#ifdef _NO_PROTO
void smTextSetInsertionLine(w, line)
Widget w;
int line;
#else
void smTextSetInsertionLine(Widget w, int line)
#endif
{
        int count_line;
        Boolean found;
        XmTextPosition pos;

        pos = 0;
        count_line = 1;
        found = True;
        while (found && count_line < line) {
            found = XmTextFindString(w, pos, "\n", XmTEXT_FORWARD, &pos);
            if (found) {
                count_line++;
                pos++;
            } /* if */
        }

        smTextSetInsertionPosition(w, pos);
}

#ifdef _NO_PROTO
Boolean is_space(ch)
char ch;
#else         /* _NO_PROTO */
Boolean is_space(char ch)
#endif
{
	static char *str = " \t";

	return (strchr(str, ch) ? True : False);
}

#ifdef _NO_PROTO
Boolean is_num(str)
char *str;
#else         /* _NO_PROTO */
Boolean is_num(char *str)
#endif
{
	Boolean isnumber = True;

	while(*str) {
	    if (!isdigit(*str)) isnumber = False;
	    str++;
	}

	return (isnumber);
}

#ifdef _NO_PROTO
Boolean stob(str)
char *str;
#else         /* _NO_PROTO */
Boolean stob(char *str)
#endif
{
	if (!strcmp(str, "True")) return (True);
	else return (False);
}

#ifdef _NO_PROTO
int search_pair(w, start, position)
Widget w;
XmTextPosition start;
XmTextPosition *position;
#else
int search_pair(Widget w, XmTextPosition start, XmTextPosition *position)
#endif
{
	int status;
	XmTextPosition i;
	char buf[3];
	int pair_matching = -1;
	char pair[2];
	Boolean search_forward = True;

        i = start;
	status = XmTextGetSubstring(w, i, 2, 3, buf);
	if (status != XmCOPY_FAILED) {
	    if (!strcmp(buf, "/*")) {
	        do {
       	            i++;
       	            status = XmTextGetSubstring(w, i, 2, 3, buf);
       	            if (status != XmCOPY_FAILED) {
       	                if (!strcmp(buf, "/*")) pair_matching--;
       	                else if (!strcmp(buf, "*/")) pair_matching++;
       	            } /* if */
       	        } while (status != XmCOPY_FAILED && pair_matching != 0);
	    }
	    else if (!strcmp(buf, "*/")) {
	        do {
                    i--;
                    status = XmTextGetSubstring(w, i, 2, 3, buf);
                    if (status != XmCOPY_FAILED) {
	                if (!strcmp(buf, "*/")) pair_matching--;
                        else if (!strcmp(buf, "/*")) pair_matching++;
                    } /* if */
                } while (i >= 0 && pair_matching != 0);
	    }
	    else if (strchr("{}()[]<>", buf[0])) {
	        switch (buf[0]) {
	            case '{' :
	                pair[0] = '{';
	                pair[1] = '}';
	                search_forward = True;
	                break;
	            case '}' :
	                pair[0] = '}';
	                pair[1] = '{';
	                search_forward = False;
	                break;
	            case '(' :
	                pair[0] = '(';
	                pair[1] = ')';
	                search_forward = True;
	                break;
	            case ')' :
	                pair[0] = ')';
	                pair[1] = '(';
	                search_forward = False;
	                break;
	            case '[' :
	                pair[0] = '[';
	                pair[1] = ']';
	                search_forward = True;
	                break;
	            case ']' :
	                pair[0] = ']';
	                pair[1] = '[';
	                search_forward = False;
	                break;
	            case '<' :
	                pair[0] = '<';
	                pair[1] = '>';
	                search_forward = True;
	                break;
	            case '>' :
	                pair[0] = '>';
	                pair[1] = '<';
	                search_forward = False;
	                break;
	        }
	        if (search_forward) {
	            do {
		        i++;
		        status = XmTextGetSubstring(w, i, 2, 3, buf);
		        if (status != XmCOPY_FAILED) {
		            if (buf[0] == pair[0]) pair_matching--;
		            else if (buf[0] == pair[1]) pair_matching++;
		        } /* if */
	            } while (i <= XmTextGetLastPosition(w) && pair_matching != 0);
	        }
	        else {
	            do {
		        i--;
		        status = XmTextGetSubstring(w, i, 2, 3, buf);
		        if (status != XmCOPY_FAILED) {
		            if (buf[0] == pair[0]) pair_matching--;
		            else if (buf[0] == pair[1]) pair_matching++;
		        } /* if */
	            } while (i >= 0 && pair_matching != 0);
	        } /* if (search_forward) */
	    } /* if (strchr("{}()[]<>", buf[0])) */
	} /* if (status != XmCOPY_FAILED) */

	if (pair_matching == 0)
	    *position = i;
	else *position = -1;

	return (pair_matching);
}

#ifdef _NO_PROTO
int find_matching(w, start, str, position)
Widget w;
XmTextPosition start;
char *str;
XmTextPosition *position;
#else
int find_matching(Widget w, XmTextPosition start, char *str, XmTextPosition *position)
#endif
{
	int status;
	XmTextPosition i = start;
	char buf[3];
	int pair_matching = -1, lineNum;
	char pair[2];
	Boolean search_forward = True;

        if (strchr("{}()[]<>", *str)) {
            switch (*str) {
                case '{' :
                    pair[0] = '{';
                    pair[1] = '}';
                    search_forward = True;
                    break;
                case '}' :
                    pair[0] = '}';
                    pair[1] = '{';
                    search_forward = False;
                    break;
                case '(' :
                    pair[0] = '(';
                    pair[1] = ')';
                    search_forward = True;
                    break;
                case ')' :
                    pair[0] = ')';
                    pair[1] = '(';
                    search_forward = False;
                    break;
                case '[' :
                    pair[0] = '[';
                    pair[1] = ']';
                    search_forward = True;
                    break;
                case ']' :
                    pair[0] = ']';
                    pair[1] = '[';
                    search_forward = False;
                    break;
                case '<' :
                    pair[0] = '<';
                    pair[1] = '>';
                    search_forward = True;
                    break;
                case '>' :
                    pair[0] = '>';
                    pair[1] = '<';
                    search_forward = False;
                    break;
            }
            if (search_forward) {
                do {
                    i++;
                    lineNum = _XmTextPosToLine((XmTextWidget)w, i);
                    if (lineNum != NOLINE) {
                        status = XmTextGetSubstring(w, i - 1, 2, 3, buf);
                        if (status != XmCOPY_FAILED) {
                            if (buf[0] == pair[0]) pair_matching--;
                            else if (buf[0] == pair[1]) pair_matching++;
                        } /* if */
                    } /* if */
                } while (lineNum != NOLINE && i < XmTextGetLastPosition(w) &&
                            pair_matching != 0);
                if (pair_matching == 0)
                    *position = i - 1;
                else *position = -1;
            }
            else {
                do {
                    i--;
                    lineNum = _XmTextPosToLine((XmTextWidget)w, i);
                    status = XmTextGetSubstring(w, i, 2, 3, buf);
                    if (status != XmCOPY_FAILED) {
                        if (buf[0] == pair[0]) pair_matching--;
                        else if (buf[0] == pair[1]) pair_matching++;
                    } /* if */
                } while (i >= 0 && pair_matching != 0);
                if (pair_matching == 0 && lineNum != NOLINE)
                    *position = i;
                else *position = -1;
            } /* if (search_forward) */
        } /* if (strchr("{}()[]<>", buf[0])) */

	return (pair_matching);
}

#ifdef _NO_PROTO
char *optimize_str(source)
char *source;
#else         /* _NO_PROTO */
char *optimize_str(char *source)
#endif
{
        char *dest;
        int dest_i = 0, sour_i = 0;
        int len;

        len = strlen(source);
        dest = XtNewString(source);
        while (*(source + sour_i) != '\0') {
            *(dest + dest_i++) = *(source + sour_i++);
            if (*(source + sour_i) == '\n' || sour_i == len)
                while (is_space(*(dest + dest_i - 1)) && dest_i > 0)
                    dest_i--;
        }
        *(dest + dest_i) = '\0';

        return (dest);
}

#ifdef _NO_PROTO
XmTextPosition smFindString(search_str, search_len, buf, buf_size,
	startpos, endpos,
	forward, case_sensitive, whole_words_only)
char *search_str;
XmTextPosition search_len;
char *buf;
XmTextPosition buf_size;
XmTextPosition startpos;
XmTextPosition endpos;
Boolean forward;
Boolean case_sensitive;
Boolean whole_words_only;
#else  /* _NO_PROTO */
XmTextPosition smFindString(char *search_str, XmTextPosition search_len,
	char *buf, XmTextPosition buf_size,
	XmTextPosition startpos, XmTextPosition endpos,
	Boolean forward, Boolean case_sensitive, Boolean whole_words_only)
#endif
{
	XmTextPosition pos = -1L;
        XmTextPosition i, n;
        char *s1, *s2;
        int  c1, c2;

        if (search_len <= 0L) return (-3L);
        if (buf_size <= 0L) return (-2L);

        if (startpos < 0L) {
	    if (forward) startpos = 0L;
	    else startpos = buf_size;
        }
        if (endpos < 0L) {
	    if (forward) endpos = buf_size;
	    else endpos = 0L;
        }

        if (endpos > buf_size) endpos = buf_size;
        if (startpos > buf_size) startpos = buf_size;

        if (case_sensitive == True) {
            if (forward) {
	        for (i = startpos; i < endpos; i++) {
	            n = search_len;
	            s1 = &buf[i];
	            s2 = search_str;
	            while (--n >= 0L && *s1++ == *s2++);
	            if (n < 0L) {
                        pos = i;
	                if (whole_words_only) {
                            if ((i > 0L && (isalnum(buf[i - 1]) || buf[i - 1] == '_')) ||
                                    (i + search_len < buf_size && (isalnum(buf[i + search_len]) ||
                                    buf[i + search_len] == '_') ))
                                pos = -1;
                            else break;
                        }
                        else break;
	            } /* if */
	        } /* for */
            }
            else { /* backward */
                for (i = startpos - 1; i >= endpos; i--) {
	            n = search_len;
	            s1 = &buf[i];
	            s2 = search_str;
	            while (--n >= 0L && *s1++ == *s2++);
	            if (n < 0L) {
		        pos = i;
		        if (whole_words_only) {
                            if ((i > 0L && (isalnum(buf[i - 1]) || buf[i - 1] == '_')) ||
                                    (i + search_len < buf_size && (isalnum(buf[i + search_len]) ||
                                    buf[i + search_len] == '_') ) )
                                pos = -1;
                            else break;
		        }
		        else break;
	            } /* if */
	        } /* for */
            }
        }
        else {	/* not case-sensitive ... */
            if (forward) {
	        for (i = startpos; i < endpos; i++) {
	            n = search_len;
	            s1 = &buf[i];
	            s2 = search_str;
	            c1 = toupper(*s1);
	            c2 = toupper(*s2);
	            while (--n >= 0L && c1 == c2) {
	                c1 = toupper(*(++s1));
	                c2 = toupper(*(++s2));
                    } /* while */
	            if (n < 0L) {
                        pos = i;
                        if (whole_words_only) {
	                    if ((i > 0L && (isalnum(buf[i - 1]) || buf[i - 1] == '_')) ||
                                    (i + search_len < buf_size && (isalnum(buf[i + search_len]) ||
                                    buf[i + search_len] == '_') ))
                                pos = -1;
                            else break;
                        }
                        else break;
	            } /* if */
	        } /* for */
            }
            else { /* backsard */
                for (i = startpos - 1; i >= endpos; i--) {
	            n = search_len;
	            s1 = &buf[i];
	            s2 = search_str;
	            c1 = toupper(*s1);
	            c2 = toupper(*s2);
	            while (--n >= 0L && c1 == c2) {
		        c1 = toupper( *(++s1));
		        c2 = toupper( *(++s2));
	            } /* while */
	            if (n < 0L) {
		        pos = i;
		        if (whole_words_only) {
		            if ((i > 0L && (isalnum(buf[i - 1]) || buf[i - 1] == '_')) ||
	                            (i + search_len < buf_size && (isalnum(buf[i + search_len]) ||
	                            buf[i + search_len] == '_') ) )
		                pos = -1;
                            else break;
		        }
		        else break;
	            } /* if */
	        } /* for */
            }
        }

        return (pos);
}

#ifdef _NO_PROTO
int shiftBlockRight(w, start, end, size)
Widget w;
XmTextPosition start;
XmTextPosition end;
int size;
#else
int shiftBlockRight(Widget w, XmTextPosition start, XmTextPosition end, int size)
#endif
{
        char insert_str[BUFSIZ];
        int i, length;
        int status;
        Boolean found;
        XmTextPosition line_start, line_end;
        char *buffer;

        for (i = 0; i < size; i++)
            insert_str[i] = ' ';
        insert_str[size] = '\0';

        length = 0;
        line_start = start;
        while (end + length > line_start) {
            found = XmTextFindString(w, line_start, "\n", XmTEXT_FORWARD, &line_end);
            if (!found) line_end = XmTextGetLastPosition(w);

            buffer = XtMalloc((Cardinal)(line_end - line_start + 1));
            status = XmTextGetSubstring(w, line_start,
                        (int)(line_end - line_start),
                        (int)(line_end - line_start + 1), buffer);
            buffer[line_end - line_start] = '\0';
            if (status != XmCOPY_FAILED) {
                i = 0;
                while (is_space(buffer[i]) && i < line_end - line_start)
                    i++;
                if (buffer[0] != '\n' && buffer[0] != '\0') {
                    XmTextInsert(w, line_start + i, insert_str);
                    length += strlen(insert_str);
                    line_end += strlen(insert_str);
                }
            }
            line_start = line_end + 1;
            XtFree(buffer);
        }

        return (length);
}

#ifdef _NO_PROTO
int shiftBlockLeft(w, start, end, size, tab_size)
Widget w;
XmTextPosition start;
XmTextPosition end;
int size;
int tab_size;
#else
int shiftBlockLeft(Widget w, XmTextPosition start, XmTextPosition end, int size, int tab_size)
#endif
{
        char *insert_str;
        int i, length, col, pos;
        int status;
        Boolean found;
        XmTextPosition line_start, line_end;
        char *buffer;

        length = 0;
        line_start = start;
        while (end + length > line_start) {
            found = XmTextFindString(w, line_start, "\n", XmTEXT_FORWARD, &line_end);
            if (!found) line_end = XmTextGetLastPosition(w);

            buffer = XtMalloc((Cardinal)(line_end - line_start + 1));
            status = XmTextGetSubstring(w, line_start,
                        (int)(line_end - line_start),
                        (int)(line_end - line_start + 1), buffer);
            buffer[line_end - line_start] = '\0';
            if (status != XmCOPY_FAILED) {
                col = 1;
                pos = 0;
                while (is_space(buffer[pos]) && pos < line_end - line_start) {
                    if (buffer[pos] == '\t')
                        col = ((col + tab_size - 1) / tab_size) * tab_size + 1;
                    else col++;
                    pos++;
                }
                if (buffer[0] != '\n' && buffer[0] != '\0') {
                    if (col - size - 1 >= 0) {
                        insert_str = XtMalloc((Cardinal)col - size);
                        for (i = 0; i < col - size - 1; i++)
                            insert_str[i] = ' ';
                        insert_str[i] = '\0';
                    }
                    else {
                        insert_str = XtMalloc(1);
                        insert_str[0] = '\0';
                    }
                    smTextReplace(w, line_start, line_start + pos, insert_str);
                    length += strlen(insert_str) - pos;
                    line_end += strlen(insert_str) - pos;
                    XtFree(insert_str);
                }
                line_start = line_end + 1;
                XtFree(buffer);
            }
        }

        return (length);
}

#ifdef _NO_PROTO
void shiftBlock(w, dir, size, tab_size)
Widget w;
int dir;
int size;
int tab_size;
#else
void shiftBlock(Widget w, int dir, int size, int tab_size)
#endif
{
        XmTextPosition start, end;
        Boolean is_selected, found;
        int length;

        is_selected = XmTextGetSelectionPosition(w, &start, &end);
        if (is_selected && start != end) {
            XmTextClearSelection(w, CurrentTime);
            if (start > 0) {
                found = XmTextFindString(w, start - 1, "\n", XmTEXT_BACKWARD, &start);
	        if (found) start++;
	        else start = 0;
	    }
	    else start = 0;
	    found = XmTextFindString(w, end - 1, "\n", XmTEXT_FORWARD, &end);
	    if (!found) end = XmTextGetLastPosition(w);

            if (dir == SHIFT_RIGHT)
                length = shiftBlockRight(w, start, end, size);
            else length = shiftBlockLeft(w, start, end, size, tab_size);

            if (found) length++;
            XmTextSetSelection(w, start, end + length, CurrentTime);
       	}
       	else {
       	    start = end = XmTextGetInsertionPosition(w);
       	    if (start > 0) {
                found = XmTextFindString(w, start - 1, "\n", XmTEXT_BACKWARD, &start);
	        if (found) start++;
	        else start = 0;
	    }
	    else start = 0;
	    found = XmTextFindString(w, end, "\n", XmTEXT_FORWARD, &end);
	    if (!found) end = XmTextGetLastPosition(w);

            if (dir == SHIFT_RIGHT)
                length = shiftBlockRight(w, start, end, size);
            else length = shiftBlockLeft(w, start, end, size, tab_size);
       	}
}

#ifdef _NO_PROTO
void UpperCase(w)
Widget w;
#else
void UpperCase(Widget w)
#endif
{
	XmTextPosition start, end;
	Boolean is_selected;
	char *str, *i_str;

	is_selected = XmTextGetSelectionPosition(w, &start, &end);
	if (is_selected && start != end) {
	    if ((str = XmTextGetSelection(w)) != NULL) {
	        i_str = str;
	        while (*i_str) {
	            if (islower(*i_str))
	            	*i_str = toupper(*i_str);
	            i_str++;
	        }
	        smTextReplace(w, start, end, str);
	    }
	}
}

#ifdef _NO_PROTO
void LowerCase(w)
Widget w;
#else
void LowerCase(Widget w)
#endif
{
	XmTextPosition start, end;
	Boolean is_selected;
	char *str, *i_str;

	is_selected = XmTextGetSelectionPosition(w, &start, &end);
	if (is_selected && start != end) {
	    if ((str = XmTextGetSelection(w)) != NULL) {
	        i_str = str;
	        while (*i_str) {
	            if (isupper(*i_str))
	            	*i_str = tolower(*i_str);
	            i_str++;
	        }
	        smTextReplace(w, start, end, str);
	    }
	}
}
